//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	attfilter.cc --- ATTitude data FILTERing
//
//	DESCRIPTION:
//		attfilter performs exposure-specific attitude data filtering and
//		GTI creation. Given a user-supplied <instexpid> string the
//		corresponding matching exposure from the current ODF (pointed at via
//		SAS_ODF	environment variable) is queried for its start and end time.
//		With this information the task is filtering the attitude data
//		time series as created by 'atthkgen' and computes the median
//		of the S/C attitude (Ra/Dec of STR viewing direction + PA)
//		on the restricted new data set. Finally, a GTI-table is created
//		under the control of a user-configurable filtering selection
//		expression which involves the previously computed median values.
//		By default the expression is doing a cone-filtering with
//		the median pointing direction as the symmetry axis:
//			skyvector(RA, DEC) in cone(RA_MED, DEC_MED, alpha)
//		alpha is the half-cone opening angle of the cone to be specified
//		via a task parameter.
//
//		The output of attfilter is a single GTI table in new or
//		existing data set. In the latter case the table shall be appended
//		to the existing set.
//
//	CHANGE HISTORY:
//		UL, 2002-05-16: 1.0.0: - inital version
//		UL, 2002-05-31: 1.1.0: - insexpid -> instexpid
//		UL, 2002-06-07: 1.2.0: - added parameters "mingtisize"/"maxfracshort"
//
//	AUTHOR:
//		U. Lammers, Mon May  6 15:15:45 MEST 2002
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Expanded task description of attfilter
//
//	$Header: /sasbuild/repositories/sasdev/sas/packages/attfilter/src/attfilter.cc,v 1.10 2007/06/28 10:18:59 aibarra Exp $
//
#include <cstdio>	//	tmpnam
#include <unistd.h>	//	unlink
#include <cmath>

#include "DataFilter.h"

#include "errstr.h"
#include "param.h"
#include "Xmm.h"
#include "Celestial.h"
#include "Dal.h"
#include "OalServer.h"
#include "stringutil.h"
#include "attfilter.h"


/*
 * Possible work around for Panther.
#ifdef MACOSX
 static long isnan(double value) { return __isnand(value); }
#endif
*/
using namespace std;
//
//	--------------------------( NanOp node )---------------------------------
//
extern "C" int isnan(double);
	// work around bug in cmath of gcc-3.1 on MacOSX where isnan is a marco
	// that is #undef'ed, hence the follwing line leads to a
	//	missing-isnan-declaration problem; on all other platform the extra
	//	extern-declaration is harmless


//
//	utilities
//
namespace {
	//	compute median value of given vector
	template<class Type>
	Type median(const Type *vin, unsigned N)
	{
		vector<Type> v(vin, vin+N);
		int k = v.size()/2;

		Type val;
		int i, j, l = 0, r = v.size()-1;
		while (r>l) {
			val = v[r];
			i = l-1;
			j = r;
			for (;;) {
				while (v[++i]<val);
				while (v[--j]>val);
				if (i>=j)
					break;
				swap(v[i], v[j]);
			}
			swap(v[i], v[r]);
			if (i>=k)
				r = i-1;
			if (i<=k)
				l = i+1;
		}

		return v[k];
	}
} // namespace


class SpacecraftAttitudeTimeSeries {
public:
	SpacecraftAttitudeTimeSeries(const string &setName,
								 const string &timeColName,
								 const string &raColName,
								 const string &decColName,
								 const string &apaColName) :
		_timeColName(timeColName), _raColName(raColName),
		_decColName(decColName), _apaColName(apaColName) {

		_table = dataSetServer->clone(setName, std::tmpnam(0), Dal::Modify)
									->table(0);	//	make copy of the set
	} // SpacecraftAttitudeTimeSeries()

	void cut(const STime &start, const STime &stop) {
		DataFilter attFilter(_timeColName + " in [" +
							 toString(start-Xmm::missionReferenceTime) + ':' +
							 toString(stop-Xmm::missionReferenceTime) + ']');
		attFilter.doFilter(*_table);
	} // cut()	

	SpacecraftAttitude medianAttitude(const bool &onlylastatt) const {
		ColumnData *racol = _table->column(_raColName)->data(),
		           *deccol = _table->column(_decColName)->data(),
			   *apacol = _table->column(_apaColName)->data();

		assert(racol->parent()->units() == "degrees");
		assert(deccol->parent()->units() == "degrees");
		assert(apacol->parent()->units() == "degrees");
		unsigned N = _table->rows();

		//If N == 0 means that there is no attitude time stamp in the exposure interval.
		//In this case, the applicaction finished properly.				

       		if (N == (unsigned) 0)
         		errstr<<"Previous filter in time results in a empty data set. Please, check exposure time"
           		<<Sas::error("EmptyDataSet");

       		real64 *racolData=0;
		real64 *deccolData=0;
		real64 *apacolData=0;
		racolData=racol->real64Data();
		deccolData=deccol->real64Data();
		apacolData=apacol->real64Data();

		double racolmed = median(racolData, N); 		  
		double deccolmed = median(deccolData, N);		   
		double apacolmed = median(apacolData, N);

		//Check that the medians are defined (eg. not NANs)
		//If this happens, take only last 100 attitude points
						
		if ( onlylastatt || isnan(racolmed) || isnan(deccolmed) || isnan(apacolmed) ) 
       		{
		 if ( isnan(racolmed) || isnan(deccolmed) || isnan(apacolmed) ) 
                 errstr<<"Median attitude out of valid region. I am taking just the last 100 attitude points for recalculation."
                 <<Sas::warning("MedianAttNotValid");
	         real64 racollast[100];
                 real64 deccollast[100];
                 real64 apacollast[100];
                 int rowCounter = N-100;
//cout << "N: " << N << endl; 
                 for (unsigned i=0 ; i<100 ; i++)
                 {
//cout << "rowCounter: " << rowCounter << endl; 
       		  racollast[i]=racolData[rowCounter];
                  deccollast[i]=deccolData[rowCounter];
                  apacollast[i]=apacolData[rowCounter];
                  rowCounter++;
                 }
                 racolData = racollast;
                 deccolData = deccollast;
                 apacolData = apacollast;
                 N = 100;
		 racolmed = median(racolData, N);  
                 deccolmed = median(deccolData, N);
                 apacolmed = median(apacolData, N);
                 if ( isnan(racolmed) || isnan(deccolmed) || isnan(apacolmed) ) 
                 errstr<<"No attitude median filtering possible. Returning."
                 <<Sas::error("MedianAttFullInvalid");                 		 
       		}
       		SpacecraftAttitude scAtt(DEG_TO_RAD*racolmed,
                    			 DEG_TO_RAD*deccolmed,
                    			 DEG_TO_RAD*apacolmed);
//		delete racolData;   
//		delete deccolData;  If I try to delete them I get segm faults!
//		delete apacolData;
		delete racol;
		delete deccol;
		delete apacol;
		return scAtt;
	} // median()

	string finalizeFilteredTable() {
		string s(_table->qualifiedName());
		_setName = _table->parent()->name();
		dataSetServer->close(_table->parent());

		return s;
	} // finalizeFilteredTable()

	~SpacecraftAttitudeTimeSeries() { unlink(_setName.c_str()); }

private:
	Table	*_table;
	string	_timeColName, _raColName, _decColName, _apaColName,
			_setName;
};


class AttitudeGtiTable {
public:
	AttitudeGtiTable(const string &gtiTableName,
					 SpacecraftAttitude &medianAttitude)
		: _gtiTableName(gtiTableName), _medianAtt(medianAttitude) {}

	//
	//	generate GTI table via tabgtigen
	//
	void generate(const string &filterType,
				  const string &userExpression,
				  const vector<string> &angles,
				  double minGtiSize,
				  const string &timeColName, const string &raColName,
				  const string &decColName, const string &apaColName,
				  const string &tableName) {

		string expr = userExpression;
		if (filterType == "cone")
			//
			//	the CONE filter
			//
			expr = "skyvector(%RA%*#RAD, %DEC%*#RAD) in "
				   "cone(skyvector(%RA_MED%, %DEC_MED%), %TOL_1%)";
		else if (filterType == "pyramid")
			//
			//	the PYRAMID filter
			//
			errstr<<"Pyramid filter not implemented yet"
				  <<Sas::error("NotYetImplemented");

		//
		//	replace symbols in selection expression
		//
		replaceall(expr, "%RA%", raColName);
		replaceall(expr, "%DEC%", decColName);
		replaceall(expr, "%APA%", apaColName);
		replaceall(expr, "%RA_MED%", toString(_medianAtt.ra(), 8));
		replaceall(expr, "%DEC_MED%", toString(_medianAtt.dec(), 8));
		replaceall(expr, "%APA_MED%", toString(_medianAtt.apos(), 8));

		for (unsigned n=0; n<angles.size(); ++n) {
			string tol = "%TOL_" + toString(n+1) + '%';
			replaceall(expr, tol, angles[n]);
		}

		//	run tabgtigen
		Info::Task* tabgtigen = task("tabgtigen");
		ParameterInfo& tabgtigenpars = *tabgtigen->parameters();
		tabgtigenpars["table"]->setValue(tableName);
		tabgtigenpars["gtiset"]->setValue(_gtiTableName);
		tabgtigenpars["expression"]->setValue(expr);
		tabgtigenpars["timecolumn"]->setValue(timeColName);
		tabgtigenpars["mingtisize"]->setValue(minGtiSize);
	
		tabgtigen->exec();
	} // generate()


	void addMedianAttributesAndCheckShortGtis(double maxFracShort,
											  double actDuration) const {
		Table *table = dataSetServer->open(_gtiTableName, Dal::Modify)
																->table(0);
		*table->addAttribute("MEDI_RA", "Right Ascension of median attitude",
							"rad") = _medianAtt.ra();
		*table->addAttribute("MEDI_DEC", "Declination  of median attitude",
							"rad") = _medianAtt.dec();
		*table->addAttribute("MEDI_APA", "Positon angke of median attitude",
							"rad") = _medianAtt.apos();

		double sumShort = table->attribute("SUMSHORT")->asReal(),
			   shortFrac = sumShort/actDuration;
		errstr<<"The sum of Good-Time-Intervals which are shorter than "
				"the minmum allowed duration of "
			  <<table->attribute("MINGTISZ")->asReal()<<"s is: "<<sumShort
			  <<"s - this is "<<setprecision(3)<<100.*shortFrac
			  <<"% of the total exposure duration. ";
		if (shortFrac > maxFracShort)
			errstr<<"Please review the tolerance angles given in "
					"parameter <tolangles>."
				  <<Sas::warning("ExcessiveNumberOfShortGtis");
		else
			errstr<<Sas::message(Msg::App, Msg::Verbose);

		dataSetServer->close(table->parent());
	} // addMedianAttributes()

private:
	string				_gtiTableName;
	SpacecraftAttitude	_medianAtt;

	void replaceall(string &s, const string &repl, const string &with) {
		string::size_type i;
		while ((i = s.find(repl)) != string::npos)
			s.replace(i, repl.length(), with);
	} // replaceall()
};


void attfilter()
{
	string timeCol = stringParameter("timecol"),
		   raCol = stringParameter("racol"),
		   decCol = stringParameter("deccol"),
		   apaCol = stringParameter("apacol");

	//	instantiate attitude time series object
	SpacecraftAttitudeTimeSeries scAttTS(stringParameter("attset"),
										 timeCol, raCol, decCol, apaCol);

	//	inquire exposure start/end times
	oalServer.instrExposureId(stringParameter("instexpid"));
	const Exposure *exp = oalServer.odf()->currentExposure();
	errstr<<"Exposure window: ["<<exp->actualStartTime()<<", "
		  <<exp->actualEndTime()<<']'
		  <<Sas::message(Msg::Meta, Msg::Silent);

	//	cut attitude TS according to exposure time window
	scAttTS.cut(exp->actualStartTime(), exp->actualEndTime());
	

	//	compute median
	SpacecraftAttitude scAtt = scAttTS.medianAttitude(booleanParameter("onlylastatt"));
	errstr<<"Median spacecraft attitude in exposure window: "
		  <<scAtt<<Sas::message(Msg::Meta, Msg::Silent);

	//	instantiate attitude GTI table ...
	AttitudeGtiTable attgti(stringParameter("gtitable"), scAtt);

	//	... trigger filtering and ...
	double maxFracShort = realParameter("maxfracshort");
	attgti.generate(stringParameter("filtertype"),
					stringParameter("expression"),
					stringListParameter("tolangles"),
					realParameter("mingtisize"),
					timeCol, raCol, decCol, apaCol,
					scAttTS.finalizeFilteredTable());

	//	... add attributes with median pointing values to GTI table
	attgti.addMedianAttributesAndCheckShortGtis(maxFracShort,
												exp->actualDuration());
} // attfilter()
